/*
Sector de arranque en GNU Assembler
Author: Erwin Meza Vega

Description:
Ver el archivo using_routines_in_real_mode.txt para mas detalles acerca del
uso de rutinas en modo real para procesadores x86.

Este bootsector ilustra el uso de rutinas en un procesador Intel x86.
Para ello, se implementan las siguientes rutinas:

* Imprimir un caracter usando los servicios de la BIOS
* Imprimir una cadena de caracteres terminada en nulo usando los servicios
  de la BIOS
* Leer una tecla usando los servicios de la BIOS

*/

.intel_syntax noprefix /* Sintaxis Intel, evitar los prefijos en los registros*/
.code16     /* 16 bits- Modo Real */

.section .text  /* Seccion de texto */

.globl start

/* Simbolo de entrada del bootsector. En este punto comienza su ejecucion. */
start:

    /* Asegurarse que cs tome el valor de 0x7C0 e ip toma el valor del
    desplazamiento de la etiqueta 'entry_point' */
    ljmp 0x7C0 : OFFSET entry_point

/* La ejecucion continua en este punto, pero se garantiza que el valor de cs
es 0x7C0*/
entry_point:

    mov ax, cs       /* Actualizar el valor de ds a partir de cs = 0x7C0 */
    mov ds, ax       /* Es necesario debido a que se van a referenciar
                           simbolos (variables) definidas en este bootsector.
                           El desplazamiento de estos simbolos se toma
                           relativo a partir del inicio del bootsector */

    /* Primero se debe configurar la pila en una region de memoria
     que no este siendo utilizada. */

     /* Ver archivo real_mode_memory.txt para los detalles de como la BIOS
     configura el primer MegaByte de memoria en modo real. */

    cli     /* Deshabilitar las interrupciones */

    mov ax, 0x9000   /* La base de la pila se configura a 0x90000 */
    mov ss, ax       /* No se puede establecer directamente ss, usar ax*/
    mov sp, 0xFFFF   /* Configurar el tope de la pila a 0x90000 + 0xFFFF*/

    sti     /* Habilitar de nuevo las interrupciones */

    /* Probar la rutina printchar */
    push '#'
    call printchar
    add sp, 2

    /* Probar la rutina print_string */
    push OFFSET mensaje
    call print_string
    add sp, 2

    /* Ahora probar la rutina para imprimir numeros hex, para mostrar
    el valor de los registros de segmento cs y ds.*/

    push cs
    call print_hex
    add sp, 2

    mov ax, cs
    cmp ax, 0x07C0
    je imprimir_menor
    jmp continuar

imprimir_menor:
    push OFFSET es_menor
    call print_string
    add sp, 2

continuar:

    push cs
    call print_hex
    add sp, 2

    jmp finished /* Saltar a el simbolo (la etiqueta) 'finished'


/* Simbolo (etiqueta) finished. */
finished:
   jmp finished   /* Salta a la etiqueta 'finished' (ciclo infinito) */

/* Implementacion de las rutinas */

/*
Rutina: printchar
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS (int 0x10)
Entrada:
  4(bp) : Los 8 bits menos significativos contienen el caracter
           ASCII a imprimir
*/
printchar:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp
    /* Almacenar el primer parametro (en la pila) en ax . Los 8 bits menos
    significativos contienen el  caracter ASCII a imprimir */
    mov ax, WORD PTR [bp + 4]


    mov ah, 0x0E     /* ah = 0x0E Imprimir un caracter en modo terminal */
    int 0x10           /* Invocar los servicios de video de la BIOS*/

    mov sp, bp       /* Cerrar el marco de pila */
    pop bp

    ret                 /* Retornar de la rutina*/

/*
Rutina: print_string
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS
Entrada:
  4(bp) : Direccion de memoria del string a imprimir en la pila
*/
print_string:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp

    /* Almacenar el primer parametro (en la pila) en si. si almacena ahora la
    direccion de memoria (el desplazamiento en bytes a partir de ds en el cual
    se encuentra el inicio de la cadena a imprimir */

    mov si, WORD PTR [bp + 4]
print_loop:
    lodsb               /* cargar un byte de ds:si en  al e incrementar si*/
    orb al, al        /* si al = 0 (caracter nulo), orb al, al =0, se
                           activa el bit ZF (zero flags) del registro FLAGS */
    jz end_print_loop   /* Si ZF esta activo, saltar al final de la rutina */

                           /* (En caso contrario, continuar la ejeucion en
                           este punto: imprimir el caracter usando
                           los servicios de video de la BIOS*/
    mov ah, 0x0E         /*Servicio para imprimir un caracter en modo texto */
    int 0x10              /* Llamada  a los servicios de video de la BIOS*/
    jmp print_loop         /* Saltar de nuevo al simbolo print_loop, para
                              imprimir el siguiente caracter de la cadena.
                              Equivalente a 'goto' en C. */

end_print_loop:         /* Parte final de la rutina */
    mov sp, bp       /* Cerrar el marco de pila creado */
    pop bp            /* Recuperar el valor de bp de la pila */
    ret            /* Retornar de la rutina */


/*
Rutina: getkey
Esta rutina lee un caracter de teclado usando los servicios de teclado
de la BIOS (int 0x16)
Entrada:
  Ninguna
Salida:
  ah contiene el Codigo de Escaneo leido
  al contiene el caracter ASCII leido

Por su simplicidad, y dado que no recibe parametros de entrada, en esta rutina
no se crea un marco de pila
*/
getkey:
    xor ah, ah   /* ah = 0, servico de video leer un caracter de teclado */
    int 0x16       /* Servicio de video */
    ret


/*
  Esta rutina imprime un numero en formato hexadecimal
  Entrada:
     4(bp): Numero a imprimir en la pila
*/
print_hex:
   push bp    /* Crear un marco de pila */
   mov bp, sp

   pusha        /* Almacenar en la pila los registros de proposito general */

   mov ax, WORD PTR [bp + 4]/* Almacenar el primer parametro de la pila en ax */

   /*
   Ahora se debe almacenar el numero pasado como parametro a la rutina (que
   se encuentra en ax) en la variable 'number'. Para ello, se utiliza la
   siguiente instruccion:
   */

   mov WORD PTR ds:[number], ax
   /*

   La instruccion anterior significa: Mover el valor del registro ax  a la
   posicion de memoria apuntada por el registro de segmento ds mas el
   desplazamiento en  bytes de la etiqueta 'number' en el bootsector.

   Dado que anteriormente se establecio ds en 0x7C0, la posicion de memoria
   a la cual se mueve el valor es:

   segmento * 16 + desplazamiento
   (0x7C0 * 0x10) + desplazamiento de la etiqueta 'number' en el bootsector

   = 0x7C00 + desplazamiento en bytes de la etiqueta 'number' en el bootsector

   = posicion de la etiqueta 'number' en memoria

   De esta forma se acceden las variables en modo real.

   Tambien es valido

    mov WORD PTR [number], ax

    ya que por defecto la instruccion mov opera sobre posiciones  de memoria
    relativas al inicio del segmento apuntado por el registro ds.

   */

   /* cx = numero de iteraciones. Se va a utilizar junto con la instruccion
    'loop' (ciclo). Esta instruccion requiere que cx contenga el numero de
    iteraciones que se desea realizar */

   mov cx, 4  /* 4 iteraciones, un numero de 16 bits tiene 4 digitos hex*/

print_digit:

  /* Tomar el valor del numero y almacenarlo en bx*/
   mov bx, WORD PTR ds:[number]

   /* Rotar el valor del registro bx 4 bits hacia la izquierda
      Con esto se logra que los 4 bits mas significativos del
      registro se conviertan en los 4 bits menos significativos.
   */
   rol bx, 4

   /* La rotacion hacia la izquierda es necesaria, debido a que el procesador
   Intel almacena los valores en formato 'little-endian', en el cual los bits
   menos significativos se almacenan primero. Por ejemplo, el numero hex
   0xABCD se almacena en memoria (y en los registros) como DCBA. Este numero
   no se puede imprimir tal cual, pues quedaria invertido.

   Para imprimir el numero 0xABCD en un formato legible se debe
   imprimir primero los digitos mas significativos: A, B, C y D.

   Por esta razon, se debe rotar hacia la izquierda el numero en 4 bits (1
   digito hexa), para tomar el digito hexa mas significativo y convertirlo
   en el digito hexa menos significativo:

   0xABCD << rotar hacia la izquierda 4 bits (1 digito hexa)
   0xBCDA
        ^
        |_ Digito que se debe imprimir : 'A'

   Repitiendo el proceso 3 veces mas,

   0xBCDA << rotar hacia la izquierda 4 bits (1 digito hexa)
   0xCDAB
        ^
        |_ Digito que se debe imprimir : 'B'

   0xCDAB << rotar hacia la izquierda 4 bits (1 digito hexa)
   0xDABC
        ^
        |_ Digito que se debe imprimir : 'C'

   0xDABC << rotar hacia la izquierda 4 bits (1 digito hexa)
   0xABCD
        ^
        |_ Digito que se debe imprimir : 'D'

   De esta forma se imprime 'ABCD' que es lo esperado.
    */

   mov WORD PTR ds:[number], bx /* almacenar el numero rotado en la variable */

   /* Extraer los primeros 4 bits del numero (un digito HEX)*/
   and bx, 0x000F

   /* La instruccion anterior toma el valor de bx (el numero con el digito
   que se desea, y de este valor se realiza una operacion logica 'and'
   con 0x000F. Con esto,  se eliminan todos los digitos hex de bx menos el
   ultimo. Por ejemplo, para el primer digito, una vez rotado el numero
   por primera vez:

   0xBCDA => bx
             and
   0x000F
   ---------  =
   0x000A => bx  Este es el valor del digito a imprimir

   */

   /* Con esto, en el registro bx se encuentra el valor numerico del digito HEX
   a imprimir. Ahora se debe convertir este valor a un caracter ASCII.*/

   mov al, BYTE PTR ds:[hex_digits + bx]

   /* La instruccion anterior significa: Mover desde el segmento apuntado por
   el registro ds, con un desplazamiento de bx, mas un desplazamiento de
   hex_digits, al registro al.

   Se debe recordar que ds:hex_digits apunta al inicio de la cadena hex_digits,
   un  string que contiene los codigos ASCII de los digitos HEX:

                    \0 (null)
                    'F'
                    'E'
                    'D'
                    'C'
                    'B'
            al  <= 'A' ds:hex_digits(bx) = 'A', si bx = A
                    '9'
                    '8'
                    '7'
                    '6'
                    '5'
                    '4'
                    '3'
                    '2'
                    '1'
  ds:hex_digits -> '0'

   */

   /*
    Ya se tiene el caracter ASCII correspondiente al digito HEX en al,
    invocar el servicio de la BIOS para imprimir un caracter
   */

   mov ah, 0x0E  /* Servicio: Imprimir un caracter*/
   int 0x10        /* Invocar los servicios de video de la BIOS */

   loop print_digit /* Iterar a la etiqueta 'print_digit'. */

   /*
     La instruccion 'loop etiqueta' permite realizar un ciclo
     de la siguiente forma:

     * decrementar cx
     * si el valor de cx es mayor que cero, saltar a la etiqueta
     * especificada en la instruccion 'loop'
     * En caso contrario, continuar con la ejecucion.
   */

   mov al, ' '   /* Imprimir un espacio al final del numero */
   int 0x10

   popa     /* Cerrar el marco de pila */
   mov sp, bp
   pop bp
   ret      /* Retornar de la rutina */


mensaje:    /* Definir un simbolo (Variable) de tipo string terminado en nulo*/
    .asciz "Hello from bootsector\r\n"

/* Variables globales de la rutina print_hex */
hex_digits: /* Cadena que contiene los digitos hex */
  .ascii "0123456789ABCDEF"
number:     /* Variable temporal para almacenar el numero  a imprimir */
  .word 0x0000

es_menor:
  .asciz "Es menor"


/*
Fin del bootsector
*/

/*

Una vez compilado, el codigo del bootsector no puede ocupar mas de 510 bytes.
Al generar el codigo ejecutable del bootsector, el linker se encarga
de rellenar con ceros el espacio a partir de la ultima instruccion de codigo
compilado hasta completar 510 bytes, y luego incluye los bytes 0x55 y 0xAA
para generar un bootsector valido de 512 bytes.

Ver el archivo link_bootsector.ld para mas detalles.

*/
