/*
Mini - Kernel en GNU Assembler
Author: Erwin Meza Vega

Descripcion:
Este kernel el leido por el bootsector (ubicado en 0x500), a la posicion
de memoria 0x1000. El bootsector le ha pasado el control al kernel por
medio de la instruccion

ljmp 0x100 : 0x0000

Con la cual cs = 0x100, ip = 0x000.

El kernel cuenta con dos secciones: seccion de texto (.text) y seccion de datos
(.data). En la seccion de texto se encuentra su codigo, y en la seccion de datos
se encuentran las variables globales.

Cada una de estas secciones se encuentran redondeadas a limites de 4KB (ver
archivo link.ld). Esto significa que que si codigo del kernel ocupa menos de
4096 bytes, el linker automaticamente redondea la seccion de texto compilado a
4096 bytes. Igual forma, si los datos del kernel (seccion .data) ocupan menos
de 4096 bytes, se redondea su seccion de texto compilado a 4096 bytes.

Por otra parte, si las secciones ocupan mas de 4096 bytes, el linker
automaticamente las redondea a limites de 4096 bytes.
*/

.intel_syntax noprefix /* Sintaxis Intel, evitar los prefijos en los registros*/
.code16     /* 16 bits- Modo Real */

.section .text  /* Seccion de texto */

.globl kernel_start

/* Simbolo de entrada del kernel. En este punto comienza su ejecucion. */
kernel_start:

    jmp kernel_main /* Saltar a el simbolo (la etiqueta) 'finished' */

.align 4
kernel_header:
.word kernel_sectors

kernel_main:

   mov ax, cs   /* Tomar el valor de cs (0x100) y pasarlo a ds */
   mov ds, ax


   /* print_string(mensaje) */
   push OFFSET mensaje
   call print_string
   add sp, 2

   /* print_string(register_message) */
   push OFFSET register_message
   call print_string
   add sp, 2

   /* print_hex(cs) */
   push cs
   call print_hex
   add sp, 2

   /* print_hex(ds) */
   push ds
   call print_hex
   add sp, 2

   /* print_hex(ss) */
   push ss
   call print_hex
   add sp, 2

   /* print_hex(es) */
   push es
   call print_hex
   add sp, 2

   /* print_hex(fs) */
   push fs
   call print_hex
   add sp, 2

   /* print_hex(gs) */
   push gs
   call print_hex
   add sp, 2

   push OFFSET dot
   call print_string
   add sp, 2


   xor ax, ax
   mov es, ax
   mov bx, 0x82
   mov ax, WORD PTR es:[bx]

   mov WORD PTR [original_ip], ax


   push WORD PTR [original_ip]
   call print_hex
   add sp, 2

   push OFFSET colon
   call print_string
   add sp, 2

   mov bx, 0x80
   mov ax, WORD PTR es:[bx]

   mov WORD PTR [original_cs], ax

   push WORD PTR [original_cs]
   call print_hex
   add sp, 2

   jmp kernel_finished

/* Simbolo kernel_finished. */
kernel_finished:
   jmp kernel_finished  /* Ciclo infinito) */


/* Implementacion de las rutinas */

/*
Rutina: printchar
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS (int 0x10)
Entrada:
  4(bp) : Los 8 bits menos significativos contienen el caracter
           ASCII a imprimir
*/
printchar:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp
    /* Almacenar el primer parametro (en la pila) en ax . Los 8 bits menos
    significativos contienen el  caracter ASCII a imprimir */
    mov ax, WORD PTR [bp + 4]

    mov ah, 0x0E     /* ah = 0x0E Imprimir un caracter en modo terminal */
    int 0x10           /* Invocar los servicios de video de la BIOS*/

    mov sp, bp       /* Cerrar el marco de pila */
    pop bp

    ret                 /* Retornar de la rutina*/

/*
Rutina: print_string
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS
Entrada:
  4(bp) : Direccion de memoria del string a imprimir en la pila
*/
print_string:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp

    /* Recuperar el primer parametro de la pila y almacenarlo en si */

    mov si, WORD PTR [bp + 4]

print_loop:
    lodsb       /* cargar un byte de ds:si en  al e incrementar si*/
    orb al, al  /* al = 0? ZF=1 */
    jz end_print_loop   /* Si ZF = 1, saltar al final de la rutina */

   /* Imprimir el caracter usando los servicios de video de la BIOS*/
    mov ah, 0x0E    /* ah = 0x0E, teletype output  */
    int 0x10
    jmp print_loop  /* iterar */

end_print_loop:     /* Parte final de la rutina */
    mov sp, bp      /* Cerrar el marco de pila */
    pop bp          /* Recuperar el valor de bp de la pila */
    ret             /* Retornar de la rutina */

/*
Rutina: getkey
Esta rutina lee un caracter de teclado usando los servicios de teclado
de la BIOS (int 0x16)
Entrada:
  Ninguna
Salida:
  ah contiene el Codigo de Escaneo leido
  al contiene el caracter ASCII leido

Por su simplicidad, y dado que no recibe parametros de entrada, en esta rutina
no se crea un marco de pila
*/
getkey:
    xor ah, ah   /* ah = 0, servico de video leer un caracter de teclado */
    int 0x16       /* Servicio de video */
    ret


/*
  Esta rutina imprime un numero en formato hexadecimal
  Entrada:
     4(bp): Numero a imprimir en la pila
*/
print_hex:
   push bp    /* Crear un marco de pila */
   mov bp, sp

   pusha        /* Almacenar en la pila los registros de proposito general */

   mov ax, WORD PTR [bp + 4]/* Almacenar el primer parametro de la pila en ax */

   /*
   number = ax
   */

   mov WORD PTR ds:[number], ax

   mov cx, 4  /* 4 iteraciones, un numero de 16 bits tiene 4 digitos hex*/

print_digit:

  /* bx = number */

   mov bx, WORD PTR ds:[number]

   /*
     Rotar bx 4 bits a la izquierda
   */
   rol bx, 4

    /* number = bx*/
   mov WORD PTR ds:[number], bx

   /* Extraer los primeros 4 bits del numero (un digito HEX)*/
   and bx, 0x000F

   /* al = hex_digits[bx] */

   mov al, BYTE PTR ds:[hex_digits + bx]

   /*
    Imprimir el caracter almacenado en al
   */

   mov ah, 0x0E  /* Servicio: Imprimir un caracter*/
   int 0x10        /* Invocar los servicios de video de la BIOS */

   loop print_digit /* Iterar a la etiqueta 'print_digit'. */

   /* fin del ciclo, imprimir un espacio al final del numero */

   mov al, ' '
   int 0x10

   popa     /* Cerrar el marco de pila */
   mov sp, bp
   pop bp
   ret      /* Retornar de la rutina */


/* Seccion de datos */

.section .data
/* Mensaje de bienvenida del kernel */
mensaje:
    .asciz "\r\nHello from real mode kernel!\r\n"

/* Mensajes para imprimir el estado de los registros de segmento */
register_message:
  .asciz "Value of segment regs:\r\ncs\t\tds\t\tss\t\tes\t\tfs\t\tgs\r\n"

/* Variables globales de la rutina print_hex */
hex_digits: /* Cadena que contiene los digitos hex */
  .ascii "0123456789ABCDEF"
number:     /* Variable temporal para almacenar el numero  a imprimir */
  .word 0x0000

dot:
.asciz ". "

colon:
.asciz ": "

original_cs:
.word 0x0000

original_ip:
.word 0x0000

/*
Fin del kernel
*/

