/*
Sector de arranque en GNU Assembler
Author: Erwin Meza Vega

Descripcion:

Este bootsector se copia a si mismo a la posicion de memoria 0x500
(la primera direccion de memoria disponible en modo real), y le cede el control
a su copia.

A su vez, esta copia del bootsector se encarga de 'leer' un kernel en formato
binario que se encuentra almacenado a partir del segundo sector de la imagen de
disco. Este kernel se carga en la direccion de memoria 0x1000 (4 KB).

El kernel almacena su tamano en sectores en una variable que se encuentra
definida exactamente en un desplazamiento de 4 bytes a partir su inicio.

El bootsector lee el primer sector del kernel (el segundo sector de la
imagen de disco), y de este primer sector obtiene el numero de sectores (-1)
que debe leer para cargar el kernel completamente. Luego, carga los demas
sectores del kernel.

Con el fin de simplificar el proceso de lectura, se implementa una rutina
que recibe como parametro el numero lineal de sector a leer, y la direccion de
memoria (segmento:offset) en el cual se debe leer.

Para cargar el kernel, se utiliza la rutina 'read_lba_from_floppy', que oculta
los detalles de la geometria de un  disco floppy.

Finalmente, el bootsector le transfiere el control al kernel por medio
de una instruccion ljmp.
*/

.intel_syntax noprefix /* Sintaxis Intel, evitar los prefijos en los registros*/
.code16     /* 16 bits- Modo Real */

.section .text  /* Seccion de texto */

.globl start

/* Simbolo de entrada del bootsector. En este punto comienza su ejecucion. */
start:

    /* Asegurarse que cs tome el valor de 0x7C0 e ip toma el valor del
    desplazamiento de la etiqueta 'entry_point' */
    ljmp 0x7C0 : OFFSET entry_point

/* La ejecucion continua en este punto, pero se garantiza que el valor de cs
es 0x7C0*/
entry_point:

   /* Copiarse a si mismo a la direccion de memoria 0x500.*/

    mov ax, cs       /* Actualizar el valor de ds a partir de cs = 0x7C0 */
    mov ds, ax       /* ds toma el mismo valor que cs, 0x7C0 */


    mov cx, 0x100 /* 0x100 = 256. El bootsector ocupa 512 bytes, pero
                        se van a copiar dos bytes (1 word) por vez.
                        Por lo tanto se deben copiar  256 words = 512 bytes. */


    xor si, si   /* si = 0. Equivalente a mov 0x0000, si */
                    /* ds:si ahora apunta a 0x7C00*/


    mov ax, 0x50
    mov es, ax   /* es = 0x50*/

    xor di, di   /* di = 0 */
                    /* es:di ahora apunta a 0x500 */

    rep movsw       /* Decrementar cx. mientras cx > 0, copiar un word de
                    ds:si a es:d e incrementar si en 2 y di en 2, ya que
                    se copio un word (2 bytes) */

   /* Pasar el control al 'clon' del bootsector, que se encuentra
     en la posicion de memoria 0x500. */

     ljmp 0x50 : OFFSET continuar /* continuar la ejecucion con cs = 0x50,
                            ip = desplazamiento de la etiqueta 'continuar'*/

    /* Este codigo es ejecutado por el 'clon' en la nueva posicion de
    memoria (0x500) */

continuar:

   /* Actualizar el registro de segmento de datos ds */

    mov ax, cs  /* Se debe actualizar ds a partir de cs, para que el */
    mov ds, ax  /* segmento de datos apunte a la nueva posicion de memoria.*/


    /* Ahora el clon debe configurar la pila en una region de memoria
     que no este siendo utilizada. */

     /* Ver archivo real_mode_memory.txt para los detalles de como la BIOS
     configura el primer MegaByte de memoria en modo real. */

    cli     /* Deshabilitar las interrupciones */

    mov ax, 0x9000   /* La base de la pila se configura a 0x90000 */
    mov ss, ax       /* No se puede establecer directamente ss, usar ax*/
    mov sp, 0xFFFF   /* Configurar el tope de la pila a 0x90000 + 0xFFFF*/

    sti     /* Habilitar de nuevo las interrupciones */


    /* Leer el primer sector del kernel */

   /* El kernel se encuentra contiguo al sector de arranque, es decir
   en el sector 2 del floppy. */

   push 0x0000
   push 0x100
   push 0x1
   /* read_lba_from_floppy(i, segmento, offset)*/
   call read_lba_from_floppy
   addw sp, 6

   /* El primer sector del kernel ha sido leido en 0x1000.
   Dentro del primer sector del kernel, a un desplazamiento de 4
   bytes, se encuentra el numero de sectores. Leer ese valor
   para leer los n-1 sectores faltantes del kernel.
   */

   /* configurar es para que apunte al segmento que comienza en 0x1000,
   la direccion de memoria  en la cual se va a leer el kernel desde disco */

   mov ax, 0x100
   mov es, ax

   mov cx, WORD PTR es:[4]
   /* es = 0x100.  es:[4] = (0x100 * 0x10) + 4 = 0x1004, la posicion
   de la variable que almacena el tamano del kernel en sectores.*/

   dec cx /* Ya se leyo el primer sector, restar 1 al numero de sectores*/

   /* Los sectores 2.. n del kernel se deben leer a partir de 0x1200, ya
   que el primer sector de 512 bytes (0x200), fue leido en 0x1000*/
   mov WORD PTR [segment], 0x120

   /* El siguiente sector a leer: 0=bootsector, 1 = primer sector del kernel
      Empezar en el sector LBA 2
   */
   mov WORD PTR [lba_sector], 0x02

read_kernel_sectors:    /* Ciclo para leer los sectores 2..n del kernel*/
   push 0x0000 /* Desplazamiento: Siempre cero */
   push WORD PTR [segment] /* Segmento: comienza en 0x120, direccion 0x1200*/
   push WORD PTR [lba_sector] /* Sector LBA a leer, comienza en 2*/
   /* Llamar a la rutina read_lba_from_floppy(i, segmento, offset=0)*/
   call read_lba_from_floppy
   add sp, 6

   inc WORD PTR [lba_sector] /* Incrementar el sector lba a leer */
   add WORD PTR [segment], 0x20 /* Incrementar el segmento en el cual se va a leer
                          el sector */

    /* Imprimir un '+' por cada sector leido. Debido a que con este ciclo
    se leen n-1 sectores del kernel, deberia imprimir n-1 caracteres '+'*/
    mov ah, 0x0E
    mov al, '+'
    int 0x10

   loop read_kernel_sectors /* Decrementar cx, e iterar si cx > 0*/


   ljmp 0x100 : 0x0000     /* Saltar a cs = 0x100, offset=0x0000 = 0x1000*/

   /* El primer sector fue leido en */

  jmp finished /* Saltar a el simbolo (la etiqueta) 'finished'


/* Simbolo (etiqueta) finished. */
finished:
   jmp finished   /* Salta a la etiqueta 'finished' (ciclo infinito) */

/* Variables del bootsector
Geometria por defecto del floppy: 80 cilindros, 2 cabezas, 18 spt.
Total sectores: 2880
Se utilizan las variables cyls, heads y spt para calcular el cilindro,
la cabeza y el sector que corresponden a un sector LBA 'i' especificado.
*/

cyls: /* Variable para almacenar los cilindros que tiene un floppy */
.word 80

heads:  /* Variable para almacenar las cabezas que tiene un floppy*/
.word 2

spt:    /* Variable para almacenar los sectores por pista de un floppy */
.word 18

segment: /* Segmento en el cual se va a leer el siguiente sector del kernel */
.word 0x0000

lba_sector: /* Numero de sector LBA del kernel a leer */
.word 0x0000

/* Implementacion de las rutinas */
/*
Rutina: read_lba_from_floppy

Descripcion: Esta rutina lee un sector de un disco floppy,
por medio de interrupciones de la BIOS.
Entrada: [bp + 4] : sector lba a leer (i).
         [bp + 6] : segmento en el cual se desea leer el sector (segmento)
         [bp + 8] : desplazamiento en el cual se desea leer el sector (offset)

LBA = Logical Block Addressing. Es una forma de referenciar un sector en el
disco, tomando el disco como una secuencia lineal de n sectores:
+--------------------------------------------------+
| 0  |   1  |   2  |  ...  |  i    | n-2     | n-1 |
+--------------------------------------------------+
En LBA, los sectores se enumeran de 0 a n-1, donde n es el
numero de sectores del disco.

Los discos generalmente utilizan una geometria basada en Cilindros, Sectores
y Cabezas (C/H/S).

A partir del sector LBA 'i' especificado, esta rutina calcula el sector, el
cilindro y la cabeza en la cual se encuentra el sector en un disco floppy.

*/

read_lba_from_floppy:
   push bp    /* Crear el marco de pila */
   mov bp, sp

   pusha       /* Guardar los registros de proposito general en la pila */

   push es   /* Guardar en la pila el registro es */

   mov ax, WORD PTR [bp + 4] /* Primer parametro: Sector (lba, comienza en 0) */
   xor dx, dx    /* dx = 0, dx:ax = sector LBA */
   div WORD PTR [spt]     /* dx:ax / spt ax = cociente, dx = resto */
   incw dx     /* Para la BIOS, los sectores empiezan en 1 */
   mov cl, dl     /* cl=Numero del sector */
   xor dx, dx    /* dx = 0
                     dx:ax = sector / spt */
   div WORD PTR [heads]       /* dx:ax (sector /spt) / cabezas ax = cociente,
                    dx = resto */

   mov dh, dl    /* dh = cabeza */
   xor dl, dl    /* Drive (0x0 = floppy) */
   mov ch, al    /* al = cilindro */

   mov bx, WORD PTR [bp + 6]
   mov es, bx    /* Segmento en el cual se va a leer el sector */
   mov bx, WORD PTR [bp + 8] /* Desplazamiento dentro del segmento en el cual se
                       va a leer el sector */

   /* es:bx apunta al segmento:desplazamiento en el cual se debe leer el
      sector */

read_sector:
   mov al, 0x1   /* Leer un (1) sector */
   mov ah, 0x2   /* ah = 02h = read sectors from drive */
   int  0x13       /* Invocar los servicios de disco de la BIOS */
   jc read_sector

   pop es        /* Recuperar el valor de es de la pila */
   popa            /* Recuperar el valor de los registros de proposito general*/
   mov sp, bp   /* Cerrar el marco de pila */
   pop bp
   ret             /* Retornar de la rutina */

/*
Fin del bootsector
*/

/*

Una vez compilado, el codigo del bootsector no puede ocupar mas de 510 bytes.
Al generar el codigo ejecutable del bootsector, el linker se encarga
de rellenar con ceros el espacio a partir de la ultima instruccion de codigo
compilado hasta completar 510 bytes, y luego incluye los bytes 0x55 y 0xAA
para generar un bootsector valido de 512 bytes.

Ver el archivo link_bootsector.ld para mas detalles.

*/
