/*
Mini - Kernel en GNU Assembler
Author: Erwin Meza Vega

Descripcion:
Este kernel el leido por el bootsector (ubicado en 0x500), a la posicion
de memoria 0x1000. El bootsector le ha pasado el control al kernel por
medio de la instruccion

ljmp 0x100 : 0x0000

Con la cual cs = 0x100, ip = 0x000.

El kernel cuenta con dos secciones: seccion de texto (.text) y seccion de datos
(.data). En la seccion de texto se encuentra su codigo, y en la seccion de datos
se encuentran las variables globales.

Cada una de estas secciones se encuentran redondeadas a limites de 4KB (ver
archivo link.ld). Esto significa que que si codigo del kernel ocupa menos de
4096 bytes, el linker automaticamente redondea la seccion de texto compilado a
4096 bytes. Igual forma, si los datos del kernel (seccion .data) ocupan menos
de 4096 bytes, se redondea su seccion de texto compilado a 4096 bytes.

Por otra parte, si las secciones ocupan mas de 4096 bytes, el linker
automaticamente las redondea a limites de 4096 bytes.
*/

.intel_syntax noprefix /* Sintaxis Intel, evitar los prefijos en los registros*/
.code16     /* 16 bits- Modo Real */

.section .text  /* Seccion de texto */

.globl kernel_start

/* Simbolo de entrada del kernel. En este punto comienza su ejecucion. */
kernel_start:

    jmp kernel_main /* Saltar a el simbolo (la etiqueta) 'finished' */

.align 4
kernel_header:
.word kernel_sectors

kernel_main:

   /* Actualizar ds a partir de cs */
   mov ax, cs
   mov ds, ax

   /*
   Configurar es:bx para que apunte a 0x80000
   */

   xor bx, bx
   mov ax, 0x8000
   mov es, ax


   /* Establecer los registros antes de invocar la interrupcion para leer
   un sector */

   /* Se va a leer el primer sector del disco (en la BIOS corresponde al sector
   1). Por esta razon los atributos C/H/S son 0/0/1. */

   /* Realizar un ciclo: Si no se puede leer el sector en 3 intentos, terminar*/
   mov WORD PTR [count], 0
read_loop:


   mov ah, 0x02 /* ah = 0x02 = Read Sectors From Drive*/
   mov al, 0x01 /* al = # de sectores  LSB cilindro */
   mov ch, 0x00 /* ch = cilindro */
   mov cl, 0x01 /* cl [0-5] = sector, [6-7] MSB cilindro */
   mov dh, 0x00 /* dh = cabeza */
   mov dl, 0x80 /* dl = drive, 0x80 = disco */


   int 0x13

  /* Se leyo un sector correctamente? */
  cmp al, 1
  je end_read_loop /* Si, romper el ciclo y continuar con la ejecucion. */

  /* No se pudo leer el sector, incrementar el numero de intentos */
  inc WORD PTR [count]

  /* comparar el numero de intentos con 3*/
  mov ax, WORD PTR [count]
  cmp ax, 3
  /* Si el numero de intentos es menor, tratar de nuevo */
  jl read_loop

  /* En caso contrario, imprimir un mensaje de error y terminar */
  push OFFSET read_error
  call print_string
  add sp, 2

  jmp kernel_finished

/* La ejecucion continua en este punto si la lectura fue correcta */
  end_read_loop:

   /* es:bx apunta al inicio del sector leido. */

   /* Apuntar al desplazamiento dentro del sector en el cual se encuentra
   la tabla de particiones: agregar 0x1BE (446)  a bx*/
   add bx, 0x1BE

   /* Empezar en una nueva linea*/
   push OFFSET enter
   call print_string
   add sp, 2

   push OFFSET partition_table_title
   call print_string
   add sp, 2

  mov cx, 4
  mov WORD PTR [count], 0
 /* Imprimir cada entrada de la tabla. */
 print_table_loop:

   push WORD PTR [count]
   call print_hex
   add sp, 2

   /* Indicador de inicio de la particion */

   xor ax, ax
   mov al, BYTE PTR es:[bx]

   push ax
   call print_hex
   add sp, 2


   xor ax, ax
   mov WORD PTR [cylinder], ax
   mov BYTE PTR [head], al
   mov BYTE PTR [sector], al

   /* Inicio de la particion (C/H/S)*/

   push OFFSET chs
   call print_string
   add sp, 2

   /* Cabeza de inicio */
   xor ax, ax
   mov al, BYTE PTR es:[bx + 1]

   mov BYTE PTR [head], al

   /* Sector de inicio */
   mov al, BYTE PTR es:[bx + 2]

   /* Poner en 0 los bits 6-7*/
   and al, 0x3F

   mov BYTE PTR [sector], al

   /* Cilindro de inicio */

   xor ax, ax
   /* Volver a sacar el sector, para tomar los 2 bits mas significativos */
   mov al, BYTE PTR es:[bx + 2]
   and al, 0xC0

   /* Desplazar a la izquierda 2 bits para que los 3 bits mas
   significativos de al queden siendo los 2 bits menos significativos de ah */
   shl ax, 2
   mov al, BYTE PTR es:[bx + 3]

   mov WORD PTR [cylinder], ax

   push WORD PTR [cylinder]
   call print_hex
   add sp, 2

   push '/'
   call printchar
   add sp, 2

   xor ah, ah
   mov al, BYTE PTR [head]

   push ax
   call print_hex
   add sp, 2

   push '/'
   call printchar
   add sp, 2

   xor ah, ah
   mov al, BYTE PTR [sector]
   push ax
   call print_hex
   add sp, 2

   /* Tipo de la particion */

   push 'T'
   call printchar
   add sp, 2

   xor ax, ax
   /* Volver a sacar el sector, para tomar los 2 bits mas significativos */
   mov al, BYTE PTR es:[bx + 4]
   push ax
   call print_hex
   add sp, 2


   /* Fin de la particion C/H/S */

   /* Cabeza de fin */
   xor ax, ax
   mov al, BYTE PTR es:[bx + 5]

   mov BYTE PTR [head], al

   /* Sector de fin */
   mov al, BYTE PTR es:[bx + 6]

   /* Poner en 0 los bits 6-7*/
   and al, 0x3F

   mov BYTE PTR [sector], al

   /* Cilindro de fin */

   xor ax, ax
   /* Volver a sacar el sector, para tomar los 2 bits mas significativos */
   mov al, BYTE PTR es:[bx + 6]
   and al, 0xC0

   /* Desplazar a la izquierda 2 bits para que los 3 bits mas
   significativos de al queden siendo los 2 bits menos significativos de ah */
   shl ax, 2
   mov al, BYTE PTR es:[bx + 7]

   mov WORD PTR [cylinder], ax

   push WORD PTR [cylinder]
   call print_hex
   add sp, 2

   push '/'
   call printchar
   add sp, 2

   xor ah, ah
   mov al, BYTE PTR [head]

   push ax
   call print_hex
   add sp, 2

   push '/'
   call printchar
   add sp, 2

   xor ah, ah
   mov al, BYTE PTR [sector]
   push ax
   call print_hex
   add sp, 2


   /* Inicio del sector en LBA
   Esta almacenado en dos words, cada  una en little-endian. */

   mov ax, WORD PTR es:[bx + 10]
   push ax
   call print_hex
   add sp, 2

   mov ax, WORD PTR es:[bx + 8]
   push ax
   call print_hex
   add sp, 2

    /* Tamanio de la particion en sectores
   Esta almacenado en dos words, cada  una en little-endian. */

   mov ax, WORD PTR es:[bx + 14]
   push ax
   call print_hex
   add sp, 2

   mov ax, WORD PTR es:[bx + 12]
   push ax
   call print_hex
   add sp, 2


   /* Pasar a la siguiente entrada de la tabla de particiones */
   add bx, 16

   /* Imprimir enter */
   push OFFSET enter
   call print_string
   add sp, 2

   /* incrementar el contador */
   inc WORD PTR [count]

   dec cx
   cmp cx, 0
   je print_table_end

   jmp print_table_loop

print_table_end:

   /* Imprimir pie de la tabla */
   push OFFSET partition_table_end
   call print_string
   add sp, 2

   call getkey

   call cls

   push '@'
   call printchar2
   add sp, 2

   jmp kernel_finished

/* Simbolo kernel_finished. */
kernel_finished:
   jmp kernel_finished  /* (Ciclo infinito) */

/* Implementacion de las rutinas */

/* Rutina: cls
Esta rutina limpia la pantalla y reinicia el apuntador en la memoria de video */
cls:
    push bp
    mov bp, sp

    pusha
    push es

    mov ax, 0xb800
    mov es, ax
    xor di, di

    mov ah, 0x07
    mov al, ' '

    mov cx, 80*25

    cls_loop:
        stosw
    loop cls_loop

    mov WORD PTR [video_offset], 0

    push WORD PTR [video_offset]
    call put_cursor
    add sp, 2

    pop es
    popa
    mov sp, bp
    pop bp
    ret


/*
Rutina: put_cursor
Esta rutina ubica el cursor en la pantalla, programando el microcontrolador CRT.
Entrada:
    [bp + 4] : Desplazamiento lineal (expresado en bytes) desde el inicio
    de la pantalla.
Salida:
    Ninguno.
*/
put_cursor:
    push bp
    mov bp, sp

    pusha
    push es

    mov ax, [bp + 4]

    mov cx, 2
    xor dx, dx

    div cx

    mov cx, ax

    mov dx, 0x3d4
    mov al, 0x0F
    outb dx, al

    mov dx, 0x3d5
    mov al, cl
    outb dx, al

    mov dx, 0x3d4
    mov al, 0x0E
    outb dx, al

    mov dx, 0x3d5
    mov al, ch
    outb dx, al


    pop es
    popa
    mov sp, bp
    pop bp
    ret

/*
Rutina: printchar
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS (int 0x10)
Entrada:
  4(bp) : Los 8 bits menos significativos contienen el caracter
           ASCII a imprimir
*/
printchar:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp
    /* Almacenar el primer parametro (en la pila) en ax . Los 8 bits menos
    significativos contienen el  caracter ASCII a imprimir */
    mov ax, WORD PTR [bp + 4]

    mov ah, 0x0E     /* ah = 0x0E Imprimir un caracter en modo terminal */
    int 0x10           /* Invocar los servicios de video de la BIOS*/

    mov sp, bp       /* Cerrar el marco de pila */
    pop bp

    ret                 /* Retornar de la rutina*/


printchar2:
    push bp
    mov bp, sp
    pusha
    push es

    mov ax, WORD PTR [bp + 4]

    cmp ax, ' '
    jg is_printable

    cmp ax, '\n'
    je newline

    cmp ax, '\r'
    je carriage

    cmp ax, '\b'
    je backspace

newline:
    jmp return_printchar2
carriage:
    jmp return_printchar2
backspace:
    dec WORD PTR [video_offset]
    mov al, ' '

is_printable:
    mov ah, 0x07
    mov di, WORD PTR [video_offset]
    mov bx, 0xb800
    mov es, bx
    stosw
    inc WORD PTR [video_offset]

return_printchar2:

    pop es
    popa

    mov sp, bp
    pop bp
    ret

/*
Rutina: print_string
Esta rutina imprime un string terminado en nulo, usando los servicios
de video de la BIOS
Entrada:
  4(bp) : Direccion de memoria del string a imprimir en la pila
*/
print_string:
    push bp           /* Crear un marco de pila para la rutina */
    mov bp, sp

    /* Recuperar el primer parametro de la pila y almacenarlo en si */

    mov si, WORD PTR [bp + 4]

print_loop:
    lodsb       /* cargar un byte de ds:si en  al e incrementar si*/
    orb al, al  /* al = 0? ZF=1 */
    jz end_print_loop   /* Si ZF = 1, saltar al final de la rutina */

   /* Imprimir el caracter usando los servicios de video de la BIOS*/
    mov ah, 0x0E    /* ah = 0x0E, teletype output  */
    int 0x10
    jmp print_loop  /* iterar */

end_print_loop:     /* Parte final de la rutina */
    mov sp, bp      /* Cerrar el marco de pila */
    pop bp          /* Recuperar el valor de bp de la pila */
    ret             /* Retornar de la rutina */

/*
Rutina: getkey
Esta rutina lee un caracter de teclado usando los servicios de teclado
de la BIOS (int 0x16)
Entrada:
  Ninguna
Salida:
  ah contiene el Codigo de Escaneo leido
  al contiene el caracter ASCII leido

Por su simplicidad, y dado que no recibe parametros de entrada, en esta rutina
no se crea un marco de pila
*/
getkey:
    xor ah, ah   /* ah = 0, servico de video leer un caracter de teclado */
    int 0x16       /* Servicio de video */
    ret


/*
  Esta rutina imprime un numero en formato hexadecimal
  Entrada:
     4(bp): Numero a imprimir en la pila
*/
print_hex:
   push bp    /* Crear un marco de pila */
   mov bp, sp

   pusha        /* Almacenar en la pila los registros de proposito general */

   mov ax, WORD PTR [bp + 4]/* Almacenar el primer parametro de la pila en ax */

   /*
   number = ax
   */

   mov WORD PTR ds:[number], ax

   mov cx, 4  /* 4 iteraciones, un numero de 16 bits tiene 4 digitos hex*/

print_digit:

  /* bx = number */

   mov bx, WORD PTR ds:[number]

   /*
     Rotar bx 4 bits a la izquierda
   */
   rol bx, 4

    /* number = bx*/
   mov WORD PTR ds:[number], bx

   /* Extraer los primeros 4 bits del numero (un digito HEX)*/
   and bx, 0x000F

   /* al = hex_digits[bx] */

   mov al, BYTE PTR ds:[hex_digits + bx]

   /*
    Imprimir el caracter almacenado en al
   */

   mov ah, 0x0E  /* Servicio: Imprimir un caracter*/
   int 0x10        /* Invocar los servicios de video de la BIOS */

   loop print_digit /* Iterar a la etiqueta 'print_digit'. */

   /* fin del ciclo, imprimir un espacio al final del numero */

   mov al, ' '
   int 0x10

   popa     /* Cerrar el marco de pila */
   mov sp, bp
   pop bp
   ret      /* Retornar de la rutina */


/* Seccion de datos */

.section .data
/* Mensaje de bienvenida del kernel */
mensaje:
    .asciz "\r\nHello from real mode kernel!\r\n"

/* Mensajes para imprimir el estado de los registros de segmento */
register_message:
  .asciz "Value of segment regs:\r\ncs\t\tds\t\tss\t\tes\t\tfs\t\tgs\r\n"

/* Variables globales de la rutina print_hex */
hex_digits: /* Cadena que contiene los digitos hex */
  .ascii "0123456789ABCDEF"
number:     /* Variable temporal para almacenar el numero  a imprimir */
  .word 0x0000

count: /* Contador usado en varios sitios del kernel */
.byte 0x00

read_error: /* Cadena de error en la lectura */
 .asciz "Error de lectura! kernel finalizado.."

partition_table_title:
.asciz "Tabla de particiones del disco duro\r\n"

partition_table_end:
.asciz "Fin de la tabla "

/* Cadena */
chs:
.asciz "C/H/S "

/* Cadena para fin de linea */
enter:
.asciz "\r\n"

/* Variable para almacenar el valor del head */
head:
.byte 0x00

/* Variable para almacenar el valor del sector */
sector:
.byte 0x00

/* Variable para almacenar el valor del cilindro */
cylinder:
.word 0x0000

/* Desplazamiento en la memoria de video */
video_offset:
.word 0x0000

/*
Fin del kernel
*/

